# -*- coding: utf-8 -*-
# Copyright (C) 2014 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS.md).

# This file is part of ianitor.

# ianitor is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# ianitor is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with ianitor.  If not, see <http://www.gnu.org/licenses/>.

from time import sleep
import signal
import logging

import consul

from ianitor.service import Service
from ianitor.args_parser import parse_args


SIGNALS = [
    member
    for member
    in dir(signal)
    if member.startswith("SIG") and '_' not in member
]


logger = logging.getLogger(__name__)


def setup_logging(verbosity):
    ilogger = logging.getLogger('ianitor')

    if verbosity:
        handler = logging.StreamHandler()
        if verbosity == 1:
            handler.setLevel(logging.ERROR)
        if verbosity == 2:
            handler.setLevel(logging.WARNING)
        if verbosity >= 3:
            handler.setLevel(logging.DEBUG)
    else:
        handler = logging.NullHandler()

    formatter = logging.Formatter(
        '[%(levelname)s] %(name)s: %(message)s'
    )

    handler.setFormatter(formatter)
    ilogger.setLevel(logging.DEBUG)
    ilogger.addHandler(handler)


def main():
    args, command = parse_args()
    setup_logging(args.verbose)

    session = consul.Consul(*args.consul_agent)

    service = Service(
        command,
        session=session,
        ttl=args.ttl,
        service_name=args.service_name,
        service_id=args.id,
        tags=args.tags,
        port=args.port
    )

    service.start()

    def signal_handler(signal_number, *_):
        service.process.send_signal(signal_number)

    for signal_name in SIGNALS:
        try:
            signum = getattr(signal, signal_name)
            signal.signal(signum, signal_handler)
        except RuntimeError:
            # signals that cannot be catched will raise RuntimeException
            # (SIGKILL) ...
            pass
        except OSError:
            # ... or OSError (SIGSTOP)
            pass

    while sleep(args.heartbeat) or service.is_up():
        service.keep_alive()

    logger.info("process quit with errorcode %s %s" % (
        service.process.poll(),
        "(signal)" if service.process.poll() and service.process.poll() < 0
        else ""
    ))

    service.deregister()



# -*- coding: utf-8 -*-
# Copyright (C) 2014 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS.md).

# This file is part of ianitor.

# ianitor is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# ianitor is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with ianitor.  If not, see <http://www.gnu.org/licenses/>.

from contextlib import contextmanager
import subprocess
import logging
from requests import ConnectionError

logger = logging.getLogger(__name__)


@contextmanager
def ignore_connection_errors(action="unknown"):
    try:
        yield
    except ConnectionError:
        logger.error("connection error on <%s> action failed" % action)


class Service(object):
    def __init__(self, command, session, ttl, service_name,
                 service_id=None, tags=None, port=None):
        self.command = command
        self.session = session
        self.process = None

        self.ttl = ttl
        self.service_name = service_name
        self.tags = tags or []
        self.port = port
        self.service_id = service_id or service_name

        self.check_id = "service:" + self.service_id

    def start(self):
        """ Start service process.

        :return:
        """
        logger.debug("starting service: %s" % " ".join(self.command))
        self.process = subprocess.Popen(self.command)
        self.register()

    def is_up(self):
        """
        Poll service process to check if service is up.

        :return:
        """
        logger.debug("polling service")
        return bool(self.process) and self.process.poll() is None

    def kill(self):
        """
        Kill service process and make sure it is deregistered from consul
        cluster.

        :return:
        """
        logger.debug("killing service")
        if self.process is None:
            raise RuntimeError("Process does not exist")

        self.process.kill()
        self.deregister()

    def register(self):
        """
        Register service in consul cluster.

        :return: None
        """
        logger.debug("registering service")
        with ignore_connection_errors():
            self.session.agent.service.register(
                name=self.service_name,
                service_id=self.service_id,
                port=self.port,
                tags=self.tags,
                # format it into XXXs format
                ttl="%ss" % self.ttl,
            )

    def deregister(self):
        """
        Deregister service from consul cluster.

        :return: None
        """
        logger.debug("deregistering service")

        with ignore_connection_errors("deregister"):
            self.session.agent.service.deregister(self.service_id)

    def keep_alive(self):
        """
        Keep alive service in consul cluster marking TTL check pass
        on consul agent.

        If some cases it can happen that service registry disappeared from
        consul cluster. This method registers service again if it happens.

        :return: None
        """
        with ignore_connection_errors("ttl_pass"):
            if not self.session.health.check.ttl_pass(self.check_id):
                # register and ttl_pass again if it failed
                logger.warning("service keep-alive failed, re-registering")
                self.register()
                self.session.health.check.ttl_pass(self.check_id)

    def __del__(self):
        """
        Cleanup processes on del
        """
        if self.process and self.process.poll() is None:
            self.kill()



# -*- coding: utf-8 -*-
# Copyright (C) 2014 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS.md).

# This file is part of ianitor.

# ianitor is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# ianitor is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with ianitor.  If not, see <http://www.gnu.org/licenses/>.

import sys
import argparse
import logging

logger = logging.getLogger(__name__)

DEFAULT_CONSUL_HTTP_API_PORT = 8500
DEFAULT_TTL = 10


class CustomFormatter(argparse.HelpFormatter):
    def __init__(self, prog):
        # default max_help_position increased for readability
        super(CustomFormatter, self).__init__(prog, max_help_position=50)

    def _format_action_invocation(self, action):
        """
        Hack _format_action_invocation to display metavar for only one
        of options
        """
        if not action.option_strings:
            metavar, = self._metavar_formatter(action, action.dest)(1)
            return metavar

        parts = []

        # if the Optional doesn't take a value, format is:
        #    -s, --long
        if action.nargs == 0:
            parts.extend(action.option_strings)

        # if the Optional takes a value, format is:
        #    -s ARGS, --long ARGS
        else:
            default = action.dest.upper()
            args_string = self._format_args(action, default)

            # here is the hack: do not add args to first option part
            # if it is both long and short
            if len(action.option_strings) > 1:
                parts.append(action.option_strings[0] + "")
                remaining = action.option_strings[1:]
            else:
                remaining = action.option_strings

            for option_string in remaining:
                parts.append('%s=%s' % (option_string, args_string))

        return ', '.join(parts)

    def add_usage(self, usage, actions, groups, prefix=None):
        """
        Hack add_usage to add fake "-- command [arguments]" to usage
        """
        actions.append(argparse._StoreAction(
            option_strings=[],
            dest="-- command [arguments]"
        ))
        return super(CustomFormatter, self).add_usage(
            usage, actions, groups, prefix
        )


def coordinates(coordinates_string):
    """ parse coordinates string
    :param coordinates_string: string in "hostname" or "hostname:port" format
    :return: (hostname, port) two-tuple
    """
    if ':' in coordinates_string:
        try:
            hostname, port = coordinates_string.split(":")
            port = int(port)

            if not hostname:
                raise ValueError()

        except ValueError:
            raise ValueError("Coordinate should be hostname or hostname:port ")
    else:
        hostname = coordinates_string
        port = DEFAULT_CONSUL_HTTP_API_PORT

    return hostname, port


def get_parser():
    """ Create ianotor argument parser with a set of reasonable defaults
    :return: argument parser
    """
    parser = argparse.ArgumentParser(
        "ianitor",
        description="Doorkeeper for consul discovered services.",
        formatter_class=CustomFormatter,
    )

    parser.add_argument(
        "--consul-agent",
        metavar="hostname[:port]", type=coordinates, default="localhost",
        help="set consul agent address"
    )

    parser.add_argument(
        "--ttl",
        metavar="seconds", type=float, default=DEFAULT_TTL,
        help="set TTL of service in consul cluster"
    )

    parser.add_argument(
        "--heartbeat",
        metavar="seconds", type=float, default=None,
        help="set process poll heartbeat (defaults to ttl/10)",
    )

    parser.add_argument(
        "--tags",
        action="append", metavar="tag",
        help="set service tags in consul cluster (can be used multiple times)",
    )

    parser.add_argument(
        "--id",
        help="set service id - must be node unique (defaults to service name)"
    )

    parser.add_argument(
        "--port",
        help="set service port",
    )

    parser.add_argument(
        "-v", "--verbose",
        action="count",
        help="enable logging to stdout (use multiple times to increase verbosity)",  # noqa
    )

    parser.add_argument(
        metavar="service-name",
        dest="service_name",
        help="service name in consul cluster",
    )

    return parser


def parse_args():
    """
    Parse program arguments.

    This function ensures that argv arguments after '--' won't be parsed by
    `argparse` and will be returned as separate list.

    :return: (args, command) two-tuple
    """

    parser = get_parser()

    try:
        split_point = sys.argv.index('--')

    except ValueError:
        if "--help" in sys.argv or "-h" in sys.argv or len(sys.argv) == 1:
            parser.print_help()
            exit(0)
        else:
            parser.print_usage()
            print(parser.prog, ": error: command missing")
            exit(1)

    else:
        argv = sys.argv[1:split_point]
        invocation = sys.argv[split_point + 1:]

        args = parser.parse_args(argv)

        # set default heartbeat to ttl / 10. if not specified
        if not args.heartbeat:
            args.heartbeat = args.ttl / 10.
            logger.debug(
                "heartbeat not specified, setting to %s" % args.heartbeat
            )

        return args, invocation
